from typing import Dict, Union
from pathlib import Path
import re
import numpy as np
import matplotlib.pyplot as plt


def _natural_sort(list_to_sort: list) -> list:
    """
    Sort list account for lack of leading zeroes.
    """
    convert = (
        lambda text: int(text) if text.isdigit() else text.lower()
    )  # noqa: E731
    alphanum_key = lambda key: [  # noqa: E731
        convert(c) for c in re.split('([0-9]+)', key)
    ]
    return sorted(list_to_sort, key=alphanum_key)


def read_elastix_iteration_data(
    iteration_txt: Union[Path, str]
) -> Dict[str, np.ndarray]:
    """
    Read and parse elastix iteration info.
    Parameters
    ----------
    iteration_txt: str or Path
        File path to an elastix iteration info file

    Returns
    -------
    iteration_dict: Dict[str, np.ndarray]
        dict containing keys of the data in the information: iteration, metric, time, etc.
    """
    with open(iteration_txt, "r") as f:
        iteration_data = f.readlines()

    iteration_data = [it.split("\t") for it in iteration_data[1:]]

    iteration_dict = {
        "iteration": np.array([int(it[0]) for it in iteration_data]),
        "metric": np.array([float(it[1]) for it in iteration_data]),
        "time[a]": np.array([float(it[2]) for it in iteration_data]),
        "step_size": np.array([float(it[3]) for it in iteration_data]),
        "gradient": np.array([float(it[4]) for it in iteration_data]),
        "iter_time": np.array([float(it[5]) for it in iteration_data]),
    }

    return iteration_dict


def read_elastix_iteration_dir(
    registration_dir: Union[Path, str]
) -> Dict[int, Dict[int, Dict[str, np.ndarray]]]:
    """
    Read a directory of iteration info and save organize it in a dict for access and plotting.

    Parameters
    ----------
    registration_dir: str or Path
        output directory of the elastix registration data

    Returns
    -------
    all_iteration_data: Dict[int, Dict[int, Dict[str, np.ndarray]]]
        Data for each registration. Keys are 0, 1, 2 for first transform, second transform, etc.
        sub-keys of each top level key are resolution 0, 1, 2, etc.

    """

    iter_info_fps = sorted(Path(registration_dir).glob("IterationInfo*"))
    iter_info_fps = [
        Path(fp) for fp in _natural_sort([str(fp) for fp in iter_info_fps])
    ]

    all_iteration_data = dict()

    for iter_fp in iter_info_fps:
        model_idx = int(iter_fp.name.split(".")[1])
        if model_idx not in all_iteration_data.keys():
            all_iteration_data.update({model_idx: dict()})
        res_idx = int(iter_fp.name.split(".")[2].strip("R"))
        model_res_data = read_elastix_iteration_data(iter_fp)
        all_iteration_data[model_idx].update({res_idx: model_res_data})

    return all_iteration_data


def read_elastix_intermediate_transform_data(
    transform_txt: Union[Path, str]
) -> Dict[str, str]:
    """
    Read transformation data into a dict for each intermediate transform.
    Parameters
    ----------
    transform_txt: str or Path
        file path to the transform parameters file

    Returns
    -------
    elastix_transform_data: Dict[str,str]
        Transform parameters for each transform in the sequence

    """
    with open(transform_txt, "r") as f:
        transform_parameters = f.readlines()

    elastix_transform_data = dict()
    for param in transform_parameters:
        if param[0] == "/":
            continue
        if param[0] == "\n":
            continue
        param = (
            param.replace("(", "")
            .replace(")", "")
            .replace("\n", "")
            .replace('"', "")
        )
        param_name, param_vals = param.split(" ", 1)
        elastix_transform_data.update({param_name: param_vals})

    return elastix_transform_data


def read_elastix_transform_dir(
    registration_dir: Union[Path, str]
) -> Dict[int, Dict[int, Dict[str, str]]]:
    """
    Read an elastix output directory's transformation data.

    Parameters
    ----------
    registration_dir: str or Path
        file path to the elastix output directory

    Returns
    -------
    all_tform_data: Dict[int, Dict[int, Dict[str, str]]]
        Transform paramteter data for each registration. Keys are 0, 1, 2 for first transform, second transform, etc.
        sub-keys of each top level key are resolution 0, 1, 2, etc.

    """
    tform_info_fps = sorted(
        Path(registration_dir).glob("TransformParameters*")
    )
    tform_info_fps = [
        Path(fp) for fp in _natural_sort([str(fp) for fp in tform_info_fps])
    ]

    all_tform_data = dict()

    for tform_fp in tform_info_fps:
        if len(tform_fp.name.split(".")) > 3:
            model_idx = int(tform_fp.name.split(".")[1])
            if model_idx not in all_tform_data.keys():
                all_tform_data.update({model_idx: dict()})
            res_idx = int(tform_fp.name.split(".")[2].strip("R"))
            model_res_data = read_elastix_intermediate_transform_data(tform_fp)
            all_tform_data[model_idx].update({res_idx: model_res_data})

    return all_tform_data


def create_iteration_plot(
    iteration_dict: Dict[str, np.ndarray], plot_title: str
) -> plt.Figure:
    """
    Generate a multi-panel plot of the elastix iteration info.

    Parameters
    ----------
    iteration_dict: Dict[str, np.ndarray]
        dict containing keys of the data in the information: iteration, metric, time, etc.
    plot_title: str
        Main title of the plot

    Returns
    -------
    fig: plt.Figure
        Matplotlib figure object on the multi-panel plot

    """
    fig, ((plt1, plt2), (plt3, plt4), (plt5, plt6)) = plt.subplots(
        3, 2, sharex=True, figsize=(8, 6)
    )
    fig.suptitle(plot_title)
    fig.subplots_adjust(hspace=0.3, wspace=0.3)
    plt_pos = {
        "metric": plt1,
        "time[a]": plt2,
        "step_size": plt3,
        "gradient": plt4,
        "iter_time": plt5,
    }

    for k, v in iteration_dict.items():
        if k == "iteration":
            continue
        x_data = iteration_dict["iteration"]
        y_data = v
        x_name = "iteration"
        y_name = k
        plt_pos[y_name].plot(x_data, y_data)
        plt_pos[y_name].set(xlabel=x_name, ylabel=y_name)
        if k in ["step_size", "gradient"]:
            plt_pos[y_name].set_yscale('log')

    plt6.axis('off')

    return fig


def write_iteration_plots(
    all_iteration_data: Dict[int, Dict[int, Dict[str, np.ndarray]]],
    data_key: str,
    output_dir: Union[str, Path],
) -> None:
    """
    Write all plots for an elastix output folder to a folder.

    Parameters
    ----------
    all_iteration_data: Dict[int, Dict[int, Dict[str, np.ndarray]]]
        Data for each registration. Keys are 0, 1, 2 for first transform, second transform, etc.
        sub-keys of each top level key are resolution 0, 1, 2, etc.
    data_key: str
        bit of text indicating which images are registered
    output_dir: str or Path
        Output path of the registration

    Returns
    -------
    None

    """
    for model_idx, model_data in all_iteration_data.items():
        for res_idx, res_data in model_data.items():
            plot_title = f"Registration info for {data_key}- transform idx {model_idx} - resolution {res_idx}"
            output_filepath = (
                Path(output_dir) / f"IterationPlot.{model_idx}.R{res_idx}.png"
            )
            out_fig = create_iteration_plot(res_data, plot_title)
            out_fig.savefig(str(output_filepath))
            plt.close(out_fig)



import multiprocessing
import random
import string
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import List, Optional, Tuple, Union

import dask.array as da
import numpy as np
import SimpleITK as sitk
import zarr
from tifffile import TiffWriter
from tiler import Tiler
from tqdm import tqdm
from wsireg.reg_images.reg_image import RegImage
from wsireg.reg_transforms.reg_transform_seq import RegTransformSeq
from wsireg.utils.im_utils import (
    format_channel_names,
    get_pyramid_info,
    prepare_ome_xml_str,
)
from wsireg.utils.tform_utils import ELX_TO_ITK_INTERPOLATORS


class OmeTiffTiledWriter:
    """
    Class for transforming, then writing whole slide images tile-by-tile, allowing
    memory-efficient transformation of images. The image
    to be transformed has to have a dask representation that is itself tiled because
    the writer finds the position of each write tile in fixed space in moving space
    and reads only the necessary portion of the image to
    perform the transformation.

    Tiles are stored in a temporary `zarr` store that is deleted after writing to OME-TIFF.

    Uses the Tiler library to manage "virtual tiling" for transformation.

    Parameters
    ----------
    reg_image: RegImage
        wsireg RegImage that has a dask store that is chunked in XY (typical of WSIs)
        .czi reader does not work!

    reg_transform_seq: RegTransformSeq
        wsireg registration transform sequence to be applied to the image

    tile_size: int
        Tile size of the output image

    zarr_tile_size: int
        Tile used in the zarr intermediate

    moving_tile_padding: int
        How much additional padded to pull from moving for each transformed tile
        This ensures that the interpolation is correctly performed during resampling
        Non-rigid transforms may need more spacing

    Attributes
    ----------
    reg_image: RegImage
        RegImage to be transformed
    reg_transform_seq: RegTransformSeq
        RegTransformSeq to be used in transformation
    tile_shape: tuple of ints
        Shape of OME-TIFF tiles going to disk
    zarr_tile_shape: tuple of ints
        Shape of zarr tiles going to disk temporarily
    moving_tile_padding: int
        Tile padding use at read in for interpolation
    """

    def __init__(
        self,
        reg_image: RegImage,
        reg_transform_seq: RegTransformSeq,
        tile_size: int = 512,
        zarr_tile_size: int = 2048,
        moving_tile_padding: int = 128,
    ):

        self._fixed_tile_positions: List[Tuple[int, int, int, int]] = []
        self._fixed_tile_positions_phys: List[
            Tuple[float, float, float, float]
        ] = []
        self._moving_tile_positions: List[Tuple[int, int, int, int]] = []
        self._moving_tile_positions_phys: List[
            Tuple[float, float, float, float]
        ] = []
        self._tiler: Optional[Tiler] = None

        self.reg_image: RegImage = reg_image
        self.reg_transform_seq: RegTransformSeq = reg_transform_seq
        self.tile_shape = (tile_size, tile_size)
        self.zarr_tile_shape = (zarr_tile_size, zarr_tile_size)
        self._check_dask_array_chunk_sizes(self.reg_image.dask_image)
        self.moving_tile_padding = moving_tile_padding
        self._build_transformation_tiles()

    @property
    def fixed_tile_positions(self) -> List[Tuple[int, int, int, int]]:
        """List of tile positions on the fixed image in pixels
        first np.ndarray is top-left x,y coordinate
        second np.ndarray is bottom-right x,y coordinate"""
        return self._fixed_tile_positions

    @property
    def fixed_tile_positions_phys(
        self,
    ) -> List[Tuple[float, float, float, float]]:
        """List of tile positions on the fixed image in physical coordinate space
        first np.ndarray is top-left x,y coordinate
        second np.ndarray is bottom-right x,y coordinate"""
        return self._fixed_tile_positions_phys

    @property
    def moving_tile_positions(self) -> List[Tuple[int, int, int, int]]:
        """Transformed coordinates of fixed tile positions to moving, pixels
        first np.ndarray is top-left x,y coordinate
        second np.ndarray is bottom-right x,y coordinate"""
        return self._moving_tile_positions

    @property
    def moving_tile_positions_phys(
        self,
    ) -> List[Tuple[float, float, float, float]]:
        """Transformed coordinates of fixed tile positions to moving, physical
        first np.ndarray is top-left x,y coordinate
        second np.ndarray is bottom-right x,y coordinate"""
        return self._moving_tile_positions_phys

    @property
    def tiler(self) -> Tiler:
        """Tiler instance to manage fixed output tiling from image shape."""
        return self._tiler

    def _create_tiler(self):
        """Create the Tiler instance."""
        self._tiler = Tiler(
            self.reg_transform_seq.output_size,
            self.zarr_tile_shape,
            overlap=0,
            mode="irregular",
        )

    def _check_dask_array_chunk_sizes(self, dask_image: da.Array) -> None:
        """Check if dask image has an acceptable chunk-size for tiled writing."""
        yx_chunks = (
            dask_image.chunksize[:2]
            if self.reg_image.is_rgb
            else dask_image.chunksize[1:]
        )

        if np.any(np.asarray(yx_chunks) > np.asarray(self.zarr_tile_shape)):
            raise ValueError(
                f"Dask image chunksize for image {str(self.reg_image.path)} "
                "is too large for tiled writing and effectively memory use is not "
                "compared to plane-by-plane writing."
            )

        return

    def _build_transformation_tiles(self):
        """Method to reinitialize tiler if there are changes."""
        self._create_tiler()
        self._get_fixed_tile_positions()
        self._get_fixed_tile_positions_phys()
        self._get_moving_tile_positions()

    def _get_and_clip_fixed_tile(
        self, tile_idx: int, output_size: Tuple[int, int], order=[1, 0]
    ):
        """Method to ensure tiles do not go beyond the output shape
        of the fixed target image"""
        tile_pos = self.tiler.get_tile_bbox(tile_idx)
        if tile_pos[1][0] > output_size[order[0]]:
            tile_pos[1][0] = output_size[order[0]]
        if tile_pos[1][1] > output_size[order[1]]:
            tile_pos[1][1] = output_size[order[1]]

        return tile_pos

    def _get_fixed_tile_positions(self):
        """Find the tile positions on the fixed image."""
        self._fixed_tile_positions = [
            self._get_and_clip_fixed_tile(
                i, self.reg_transform_seq.output_size, order=[0, 1]
            )
            for i in range(self.tiler.n_tiles)
        ]

    def _get_fixed_tile_positions_phys(self):
        """Fixed tile pixel indices to physical coordinates
        used in ITK transforms."""
        self._fixed_tile_positions_phys = [
            (
                f[0] * self.reg_transform_seq.output_spacing,
                f[1] * self.reg_transform_seq.output_spacing,
            )
            for f in self._fixed_tile_positions
        ]

    def _get_moving_tile_positions(self):
        """Method to transform tile positions in fixed
        to moving so that each write tile in fixed has a corresponding
        read region in moving."""
        for fixed_tile_pos in self._fixed_tile_positions_phys:
            corners_phys = []
            corners_px = []

            for idx, corner in enumerate(fixed_tile_pos):
                if idx == 0:
                    corner -= self.moving_tile_padding
                if idx == 1:
                    corner += self.moving_tile_padding
                for idx, t in enumerate(
                    self.reg_transform_seq.reg_transforms[::-1]
                ):
                    if idx == 0:
                        t_pt = t.itk_transform.TransformPoint(corner.tolist())
                    else:
                        t_pt = t.itk_transform.TransformPoint(t_pt)

                t_pt = np.array(t_pt)
                t_pt_px = t_pt / self.reg_image.image_res
                corners_phys.append(t_pt)
                corners_px.append(t_pt_px)

            self._moving_tile_positions_phys.append(tuple(corners_phys))
            self._moving_tile_positions.append(tuple(corners_px))

    def set_output_spacing(
        self, output_spacing: Tuple[Union[int, float], Union[int, float]]
    ) -> None:
        """
        Sets the output spacing of the resampled image and will change
        output shape accordingly
        Parameters
        ----------
        output_spacing: Tuple[Union[int,float], Union[int,float]]
            Spacing of grid for resampling. Will default to target image spacing
        """
        self.reg_transform_seq.set_output_spacing(output_spacing)
        self._build_transformation_tiles()

    def set_tile_size(self, tile_size: int) -> None:
        """
        Set the internal tile size of the OME-TIFF to be written.
        Parameters
        ----------
        tile_size: int
            tile size in pixels in x and y for the OME-TIFF

        """
        self.tile_shape = (tile_size, tile_size)

    def set_zarr_tile_size(self, tile_size: int) -> None:
        """
        Set the tile size for the zarr intermediate.
        Parameters
        ----------
        tile_size: int
            tile size in pixels in x and y for the temporary zarr store
        """
        self.zarr_tile_shape = (tile_size, tile_size)
        self._build_transformation_tiles()

    def _create_tile_resampler(
        self, tile_origin: Tuple[float, float]
    ) -> sitk.ResampleImageFilter:
        """
        Build each tile's resampler.
        Parameters
        ----------
        tile_origin: Tuple[float, float]
            Position of the tile in physical coordinates

        Returns
        -------
        resampler: sitk.ResampleImageFilter
            resampler for an individual fixed tile
        """
        resampler = sitk.ResampleImageFilter()
        resampler.SetOutputOrigin(tile_origin)
        resampler.SetOutputDirection(
            self.reg_transform_seq.reg_transforms[-1].output_direction
        )
        resampler.SetSize(self.zarr_tile_shape)
        resampler.SetOutputSpacing(self.reg_transform_seq.output_spacing)

        interpolator = ELX_TO_ITK_INTERPOLATORS.get(
            self.reg_transform_seq.reg_transforms[-1].resample_interpolator
        )
        resampler.SetInterpolator(interpolator)
        resampler.SetTransform(self.reg_transform_seq.composite_transform)

        return resampler

    def write_tiles_to_zarr_store(
        self,
        temp_zarr_store: zarr.TempStore,
        max_workers: Optional[int] = None,
    ):
        """
        Write tiles to a temporary zarr store.

        Parameters
        ----------
        temp_zarr_store: zarr.TempStore
            Temporary store where the dataset will go.

        Returns
        -------
        resample_zarray: zarr.Array
            zarr store contained transformed images
        """
        zgrp = zarr.open(temp_zarr_store)
        if self.reg_image.is_rgb:
            resample_zarray = zgrp.create_dataset(
                random_str(),
                shape=(
                    self.reg_transform_seq.output_size[1],
                    self.reg_transform_seq.output_size[0],
                    self.reg_image.shape[-1],
                ),
                chunks=self.tile_shape,
                dtype=self.reg_image.im_dtype,
            )
        else:
            resample_zarray = zgrp.create_dataset(
                random_str(),
                shape=(
                    self.reg_image.n_ch,
                    self.reg_transform_seq.output_size[1],
                    self.reg_transform_seq.output_size[0],
                ),
                chunks=(1,) + self.tile_shape,
                dtype=self.reg_image.im_dtype,
            )

        self._transform_write_tile_set(
            resample_zarray, max_workers=max_workers
        )

        return resample_zarray

    def _transform_write_tile(self, data):
        """Worker function to transform and place tile in zarr store."""
        (
            resample_zarray,
            ch_idx,
            fixed_tile_position,
            fixed_tile_origin,
            moving_tile_corners,
        ) = data

        tile_resampler = self._create_tile_resampler(fixed_tile_origin)

        x_size, y_size = self._get_image_size()

        x_max, x_min, y_max, y_min = self._get_moving_tile_slice(
            moving_tile_corners, x_size, y_size
        )

        tile_resampled = self._resample_tile(
            ch_idx, tile_resampler, x_max, x_min, y_max, y_min
        )

        if tile_resampled:
            (
                x_max_fixed,
                x_min_fixed,
                y_max_fixed,
                y_min_fixed,
            ) = self._get_fixed_slice(fixed_tile_position)

            x_max, y_max = self._correct_end_moving_slices(
                x_max_fixed,
                x_min_fixed,
                y_max_fixed,
                y_min_fixed,
            )

            if self.reg_image.is_rgb:
                resample_zarray[
                    y_min_fixed:y_max_fixed, x_min_fixed:x_max_fixed, :
                ] = sitk.GetArrayFromImage(tile_resampled)[:y_max, :x_max, :]
            else:
                resample_zarray[
                    ch_idx, y_min_fixed:y_max_fixed, x_min_fixed:x_max_fixed
                ] = sitk.GetArrayFromImage(tile_resampled)[:y_max, :x_max]

    def _get_image_size(self) -> Tuple[int, int]:
        """Get moving image size for tile dilineation"""
        x_size = (
            self.reg_image.shape[1]
            if self.reg_image.is_rgb
            else self.reg_image.shape[2]
        )
        y_size = (
            self.reg_image.shape[0]
            if self.reg_image.is_rgb
            else self.reg_image.shape[1]
        )
        return x_size, y_size

    def _resample_tile(
        self,
        ch_idx: int,
        tile_resampler: sitk.ResampleImageFilter,
        x_max: int,
        x_min: int,
        y_max: int,
        y_min: int,
    ) -> Optional[sitk.Image]:
        """Resample tile or don't if it is outside of the moving
        image space."""

        if x_min == 0 and x_max == 0:
            return

        if y_min == 0 and y_max == 0:
            return

        if self.reg_image.is_rgb:
            image = self.reg_image.dask_image[y_min:y_max, x_min:x_max, :]
            image = sitk.GetImageFromArray(image, isVector=True)
        elif self.reg_image.n_ch == 1:
            image = da.squeeze(self.reg_image.dask_image)[
                y_min:y_max, x_min:x_max
            ]
            image = sitk.GetImageFromArray(image, isVector=False)
        else:
            image = self.reg_image.dask_image[ch_idx, y_min:y_max, x_min:x_max]
            image = sitk.GetImageFromArray(image, isVector=False)
        image.SetSpacing((self.reg_image.image_res, self.reg_image.image_res))
        image.SetOrigin(
            image.TransformIndexToPhysicalPoint([int(x_min), int(y_min)])
        )

        tile_resampled = tile_resampler.Execute(image)

        return tile_resampled

    def _correct_end_moving_slices(
        self,
        x_max_fixed: int,
        x_min_fixed: int,
        y_max_fixed: int,
        y_min_fixed: int,
    ) -> Tuple[int, int]:
        """Correct tiles that extend past the size of the fixed coordinate space."""
        # correct for end tiles
        y_max = y_max_fixed - y_min_fixed
        x_max = x_max_fixed - x_min_fixed
        return x_max, y_max

    def _get_fixed_slice(
        self, fixed_tile_position: Tuple[np.ndarray, np.ndarray]
    ) -> Tuple[int, int, int, int]:
        """Get tile slice in fixed tile pixels."""
        y_min_fixed = fixed_tile_position[0][1]
        x_min_fixed = fixed_tile_position[0][0]
        y_max_fixed = fixed_tile_position[1][1]
        x_max_fixed = fixed_tile_position[1][0]
        return x_max_fixed, x_min_fixed, y_max_fixed, y_min_fixed

    def _get_moving_tile_slice(
        self,
        moving_tile_corners: Tuple[np.ndarray, np.ndarray],
        x_size: int,
        y_size: int,
    ) -> Tuple[int, int, int, int]:
        """Get tile slice in moving tile pixels."""
        x_min = (
            moving_tile_corners[0][0] if moving_tile_corners[0][0] >= 0 else 0
        )
        x_min = x_min if x_min <= x_size else x_size
        x_min = np.ceil(x_min).astype(int)
        x_max = (
            moving_tile_corners[1][0] if moving_tile_corners[1][0] >= 0 else 0
        )
        x_max = x_max if x_max <= x_size else x_size
        x_max = np.ceil(x_max).astype(int)
        y_min = (
            moving_tile_corners[0][1] if moving_tile_corners[0][1] >= 0 else 0
        )
        y_min = y_min if y_min <= y_size else y_size
        y_min = np.ceil(y_min).astype(int)
        y_max = (
            moving_tile_corners[1][1] if moving_tile_corners[1][1] >= 0 else 0
        )
        y_max = y_max if y_max <= y_size else y_size
        y_max = np.ceil(y_max).astype(int)
        # catch changing positions of x and y when there are coordinate flips
        if y_min > y_max:
            y_max_temp = y_min
            y_min_temp = y_max
            y_max = y_max_temp
            y_min = y_min_temp
        if x_min > x_max:
            x_max_temp = x_min
            x_min_temp = x_max
            x_max = x_max_temp
            x_min = x_min_temp
        return x_max, x_min, y_max, y_min

    def _transform_write_tile_set(
        self, resample_zarray: zarr.Array, max_workers: Optional[int] = None
    ):
        """Function to loop over all channels and tile positions
        and write to zarr"""
        if max_workers == 1:
            use_multiprocessing = False
        else:
            use_multiprocessing = True
            if not max_workers:
                max_workers = multiprocessing.cpu_count()

        n_ch = 1 if self.reg_image.is_rgb else self.reg_image.n_ch
        all_tile_args = []
        for ch_idx in range(n_ch):
            for ft_pos, mt_pos in tqdm(
                zip(
                    self._fixed_tile_positions,
                    self._moving_tile_positions,
                ),
                total=len(self._fixed_tile_positions),
                desc="Writing zarr tiles",
                unit=" tile",
                disable=True if use_multiprocessing else False,
            ):
                tile_origin = (
                    ft_pos[0] * self.reg_transform_seq.output_spacing[0]
                )
                tile_args = (
                    resample_zarray,
                    ch_idx,
                    ft_pos,
                    tuple(tile_origin.astype(float)),
                    mt_pos,
                )
                all_tile_args.append(tile_args)
                if not use_multiprocessing:
                    self._transform_write_tile(tile_args)

        if use_multiprocessing:
            with ThreadPoolExecutor(max_workers) as executor:
                _ = list(
                    tqdm(
                        executor.map(
                            self._transform_write_tile, all_tile_args
                        ),
                        total=len(all_tile_args),
                        desc="Writing zarr tiles",
                        unit=" tile",
                    )
                )

    def _prepare_image_info(
        self,
        image_name,
        write_pyramid=True,
    ):
        """Prepare info for pyramidalization and create OME-TIFF."""

        x_size, y_size = self.reg_transform_seq.output_size
        x_spacing, y_spacing = self.reg_transform_seq.output_spacing

        out_tile_shape = self.tile_shape
        # protect against too large tile size
        while (
            y_size / out_tile_shape[0] <= 1 or x_size / out_tile_shape[0] <= 1
        ):
            out_tile_shape = (out_tile_shape[0] // 2, out_tile_shape[1] // 2)

        pyr_levels, _ = get_pyramid_info(
            y_size, x_size, self.reg_image.n_ch, self.tile_shape[0]
        )

        n_pyr_levels = len(pyr_levels)

        PhysicalSizeY = y_spacing
        PhysicalSizeX = x_spacing

        channel_names = format_channel_names(
            self.reg_image.channel_names, self.reg_image.n_ch
        )

        omexml = prepare_ome_xml_str(
            y_size,
            x_size,
            self.reg_image.n_ch,
            self.reg_image.im_dtype,
            self.reg_image.is_rgb,
            PhysicalSizeX=PhysicalSizeX,
            PhysicalSizeY=PhysicalSizeY,
            PhysicalSizeXUnit="µm",
            PhysicalSizeYUnit="µm",
            Name=image_name,
            Channel=None if self.reg_image.is_rgb else {"Name": channel_names},
        )

        subifds = n_pyr_levels - 1 if write_pyramid is True else None

        return n_pyr_levels, subifds, out_tile_shape, omexml

    def _transformed_tile_generator(self, d_array: da.Array, ch_idx: int):
        """Create generator of tifffile tiles for OME-TIFF."""
        out_shape = (
            d_array.shape[:2] if self.reg_image.is_rgb else d_array.shape[1:]
        )
        for y in range(0, out_shape[0], self.tile_shape[0]):
            for x in range(0, out_shape[1], self.tile_shape[1]):
                if self.reg_image.is_rgb:
                    yield d_array[
                        y : y + self.tile_shape[0],
                        x : x + self.tile_shape[1],
                        :,
                    ].compute()
                else:
                    yield d_array[
                        ch_idx,
                        y : y + self.tile_shape[0],
                        x : x + self.tile_shape[1],
                    ].compute()

    def write_image_by_tile(
        self,
        image_name: str,
        output_dir: Union[Path, str] = "",
        write_pyramid: bool = True,
        compression: Optional[str] = "default",
        zarr_temp_dir: Optional[Union[str, Path]] = None,
    ) -> str:
        """
        Write images to OME-TIFF from temp zarr store with data.

        Parameters
        ----------
        image_name: str
            file path stem of the image to be written
        output_dir: Union[str,Path]
            directory where image is to be written
        write_pyramid: bool
            whether to write a pyramid or single layer
        compression: str
            Use compression. "default" will be lossless "deflate" for non-rgb images
            and "jpeg" for RGB images
        zarr_temp_dir: Path or str
            Directory to store the temporary zarr data
            (mostly used for debugging)

        Returns
        -------
        output_file_name: Path
            Path to written image file
        """
        zstr = zarr.TempStore(dir=zarr_temp_dir)
        try:
            resample_zarray = self.write_tiles_to_zarr_store(zstr)
            output_file_name = str(Path(output_dir) / f"{image_name}.ome.tiff")

            if compression == "default":
                print("using default compression")
                compression = "jpeg" if self.reg_image.is_rgb else "deflate"
            else:
                compression = compression

            (
                n_pyr_levels,
                subifds,
                out_tile_shape,
                omexml,
            ) = self._prepare_image_info(
                image_name, write_pyramid=write_pyramid
            )

            print(f"saving to {output_file_name}")

            dask_image = da.from_zarr(resample_zarray)
            options = dict(
                tile=self.tile_shape,
                compression=compression,
                photometric="rgb" if self.reg_image.is_rgb else "minisblack",
                metadata=None,
            )
            with TiffWriter(output_file_name, bigtiff=True) as tif:
                if self.reg_image.is_rgb:
                    print(
                        f"writing base layer RGB - shape: {dask_image.shape}"
                    )
                    # tile_iterator_strides = self._get_tile_iterator_strides(dask_image)
                    tile_iterator = self._transformed_tile_generator(
                        dask_image, 0
                    )
                    tif.write(
                        tile_iterator,
                        subifds=subifds,
                        description=omexml,
                        shape=dask_image.shape,
                        dtype=dask_image.dtype,
                        **options,
                    )

                    if write_pyramid:
                        for pyr_idx in range(1, n_pyr_levels):
                            sub_res = compute_sub_res(
                                dask_image,
                                pyr_idx,
                                self.tile_shape[0],
                                self.reg_image.is_rgb,
                                self.reg_image.im_dtype,
                            )
                            print(
                                f"pyr {pyr_idx} : RGB-shape: {sub_res.shape}"
                            )

                            # tile_strides = self._get_tile_iterator_strides(sub_res)
                            sub_res_tile_iterator = (
                                self._transformed_tile_generator(sub_res, 0)
                            )
                            tif.write(
                                sub_res_tile_iterator,
                                shape=sub_res.shape,
                                dtype=self.reg_image.im_dtype,
                                **options,
                                subfiletype=1,
                            )
                else:
                    for channel_idx in range(self.reg_image.n_ch):

                        description = omexml if channel_idx == 0 else None
                        print(
                            f"writing channel {channel_idx} - shape: {dask_image.shape[1:]}"
                        )
                        tile_iterator = self._transformed_tile_generator(
                            dask_image, channel_idx
                        )

                        tif.write(
                            tile_iterator,
                            subifds=subifds,
                            description=description,
                            shape=dask_image.shape[1:],
                            dtype=dask_image.dtype,
                            **options,
                        )
                        if write_pyramid:
                            for pyr_idx in range(1, n_pyr_levels):
                                sub_res = compute_sub_res(
                                    dask_image,
                                    pyr_idx,
                                    self.tile_shape[0],
                                    self.reg_image.is_rgb,
                                    self.reg_image.im_dtype,
                                )

                                sub_res_tile_iterator = (
                                    self._transformed_tile_generator(
                                        sub_res, channel_idx
                                    )
                                )

                                tif.write(
                                    sub_res_tile_iterator,
                                    shape=sub_res.shape[1:],
                                    dtype=dask_image.dtype,
                                    **options,
                                    subfiletype=1,
                                )
            try:
                resample_zarray.store.clear()
            except FileNotFoundError:
                pass
            return output_file_name

        # bare except to always clear temporary storage on failure
        except Exception as e:
            print(e)
            try:
                resample_zarray.store.clear()
            except FileNotFoundError:
                pass


def compute_sub_res(
    zarray: da.Array,
    pyr_level: int,
    tile_size: int,
    is_rgb: bool,
    im_dtype: np.dtype,
) -> da.Array:
    """
    Compute factor-of-2 sub-resolutions from dask array for pyramidalization using dask.

    Parameters
    ----------
    zarray: da.Array
        Dask array to be downsampled
    pyr_level: int
        level of the pyramid. 0 = base, 1 = 2x downsampled, 2=4x downsampled...
    tile_size: int
        Size of tiles in dask array after downsampling
    is_rgb: bool
        whether dask array is RGB interleaved
    im_dtype: np.dtype
        dtype of the output da.Array

    Returns
    -------
    resampled_zarray_subres: da.Array
        Dask array (unprocessed) to be written
    """
    if is_rgb:
        resampling_axis = {0: 2**pyr_level, 1: 2**pyr_level, 2: 1}
        tiling = (tile_size, tile_size, 3)
    else:
        resampling_axis = {0: 1, 1: 2**pyr_level, 2: 2**pyr_level}
        tiling = (1, tile_size, tile_size)

    resampled_zarray_subres = da.coarsen(
        np.mean,
        zarray,
        resampling_axis,
        trim_excess=True,
    )
    resampled_zarray_subres = resampled_zarray_subres.astype(im_dtype)
    resampled_zarray_subres = resampled_zarray_subres.rechunk(tiling)

    return resampled_zarray_subres


def random_str() -> str:
    """Get a random string to store the zarr array"""
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for _ in range(10))



import json
from abc import ABC
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import dask.array as da
import itk
import numpy as np
import SimpleITK as sitk

from wsireg.parameter_maps.preprocessing import ImagePreproParams
from wsireg.reg_shapes import RegShapes
from wsireg.utils.im_utils import (
    compute_mask_to_bbox,
    contrast_enhance,
    sitk_inv_int,
    sitk_max_int_proj,
    transform_plane,
)
from wsireg.utils.tform_utils import (
    gen_aff_tform_flip,
    gen_rig_to_original,
    gen_rigid_tform_rot,
    gen_rigid_translation,
    prepare_wsireg_transform_data,
)


class RegImage(ABC):
    """Base class for registration images"""

    _path: Union[str, Path]

    # image data
    _dask_image: da.Array
    _reg_image: Union[sitk.Image, itk.Image]
    _mask: Optional[Union[sitk.Image, itk.Image]] = None

    # image dimension information
    _shape: Tuple[int, int, int]
    _n_ch: int
    _im_dtype: np.dtype

    # channel information
    _channel_axis: int
    _is_rgb: bool
    _is_interleaved: bool
    _channel_names: List[str]
    _channel_colors: List[str]

    # scaling information
    _image_res: Union[Tuple[int, int], Tuple[float, float]]

    # reg image preprocessing
    _preprocessing: Optional[ImagePreproParams] = None

    def __init__(
        self, preprocessing: Optional[Union[ImagePreproParams, Dict]] = None
    ):
        if preprocessing:
            if isinstance(preprocessing, ImagePreproParams):
                self._preprocessing = preprocessing
            elif isinstance(preprocessing, dict):
                self._preprocessing = ImagePreproParams(**preprocessing)
        else:
            self._preprocessing = ImagePreproParams()

    @property
    def path(self) -> Union[str, Path]:
        """Path to image file."""
        return self._path

    @property
    def shape(self) -> Tuple[int, int, int]:
        """Shape of image file (C,Y,X) or (Y,X,C) if RGB"""
        return self._shape

    @property
    def n_ch(self) -> int:
        """Number of channels in image."""
        return self._n_ch

    @property
    def im_dtype(self) -> np.dtype:
        """Data type of image"""
        return self._im_dtype

    @property
    def is_rgb(self) -> bool:
        """Whether image is RGB or not."""
        return self._is_rgb

    @property
    def is_interleaved(self) -> bool:
        """Whether RGB image is interleaved or not."""
        return self._is_interleaved

    @property
    def channel_axis(self) -> int:
        """Axis of the channel dimension."""
        return self._channel_axis

    @property
    def image_res(self) -> Union[float, int]:
        """Spacing of image pixels (only isotropic right now)"""
        return self._image_res

    @property
    def channel_names(self) -> List[str]:
        """Name of the channels of the image."""
        return self._channel_names

    @property
    def channel_colors(self) -> List[str]:
        """Colors of the channels."""
        return self._channel_colors

    @property
    def dask_image(self) -> da.Array:
        """Dask representation of the image."""
        return self._dask_image

    @property
    def mask(self) -> Optional[Union[sitk.Image, itk.Image]]:
        """Mask of the image."""
        return self._mask

    @property
    def reg_image(self) -> Union[sitk.Image, itk.Image]:
        """Preprocessed version of image for registration"""
        return self._reg_image

    @property
    def preprocessing(self) -> Optional[ImagePreproParams]:
        """Preprocessing params to make `reg_image`"""
        return self._preprocessing

    def read_mask(
        self, mask: Union[str, Path, sitk.Image, np.ndarray]
    ) -> sitk.Image:
        """
        Read a mask from geoJSON or a binary image.

        Parameters
        ----------
        mask: path to image/geoJSON or image
            Data to be used to make the mask, can be a path to a geoJSON
            or an image file, or a if an np.ndarray, used directly.

        Returns
        -------
        mask: sitk.Image
            Mask image with spacing/size of `reg_image`
        """
        if isinstance(mask, np.ndarray):
            mask = sitk.GetImageFromArray(mask)
        elif isinstance(mask, (str, Path)):
            if Path(mask).suffix.lower() == ".geojson":
                out_shape = self.shape[:2] if self.is_rgb else self.shape[1:]
                mask_shapes = RegShapes(mask)
                mask = mask_shapes.draw_mask(out_shape[::-1], labels=False)
                mask = sitk.GetImageFromArray(mask)
            else:
                mask = sitk.ReadImage(mask)
        elif isinstance(mask, sitk.Image):
            mask = mask

        mask.SetSpacing((self.image_res, self.image_res))

        return mask

    def preprocess_reg_image_intensity(
        self, image: sitk.Image, preprocessing: ImagePreproParams
    ) -> sitk.Image:
        """
        Preprocess image intensity data to single channel image.

        Parameters
        ----------
        image: sitk.Image
            reg_image to be preprocessed
        preprocessing: ImagePreproParams
            Parameters of the preprocessing

        Returns
        -------
        image: sitk.Image
            Preprocessed single-channel image
        """

        if preprocessing.image_type.value == "FL":
            preprocessing.invert_intensity = False
        elif preprocessing.image_type.value == "BF":
            preprocessing.max_int_proj = False
            preprocessing.contrast_enhance = False
            if self.is_rgb:
                preprocessing.invert_intensity = True

        if preprocessing.max_int_proj:
            image = sitk_max_int_proj(image)

        if preprocessing.contrast_enhance:
            image = contrast_enhance(image)

        if preprocessing.invert_intensity:
            image = sitk_inv_int(image)

        if preprocessing.custom_processing:
            for k, v in preprocessing.custom_processing.items():
                print(f"performing preprocessing: {k}")
                image = v(image)

        image.SetSpacing((self.image_res, self.image_res))

        return image

    def preprocess_reg_image_spatial(
        self,
        image: sitk.Image,
        preprocessing: ImagePreproParams,
        imported_transforms=None,
    ) -> Tuple[sitk.Image, List[Dict]]:
        """
        Spatial preprocessing of the reg_image.

        Parameters
        ----------
        image: sitk.Image
            reg_image to be preprocessed
        preprocessing: ImagePreproParams
            Spatial preprocessing parameters
        imported_transforms:
            Not implemented yet..

        Returns
        -------
        image: sitk.Image
            Spatially preprcessed image ready for registration
        transforms: list of transforms
            List of pre-initial transformations
        """

        transforms = []
        original_size = image.GetSize()

        if preprocessing.downsampling > 1:
            print(
                "performing downsampling by factor: {}".format(
                    preprocessing.downsampling
                )
            )
            image.SetSpacing((self.image_res, self.image_res))
            image = sitk.Shrink(
                image,
                (preprocessing.downsampling, preprocessing.downsampling),
            )

            if self._mask is not None:
                self._mask.SetSpacing((self.image_res, self.image_res))
                self._mask = sitk.Shrink(
                    self._mask,
                    (
                        preprocessing.downsampling,
                        preprocessing.downsampling,
                    ),
                )

            image_res = image.GetSpacing()[0]
        else:
            image_res = self.image_res

        if float(preprocessing.rot_cc) != 0.0:
            print(f"rotating counter-clockwise {preprocessing.rot_cc}")
            rot_tform = gen_rigid_tform_rot(
                image, image_res, preprocessing.rot_cc
            )
            (
                composite_transform,
                _,
                final_tform,
            ) = prepare_wsireg_transform_data({"initial": [rot_tform]})

            image = transform_plane(image, final_tform, composite_transform)

            if self._mask is not None:
                self._mask.SetSpacing((image_res, image_res))
                self._mask = transform_plane(
                    self._mask, final_tform, composite_transform
                )
            transforms.append(rot_tform)

        if preprocessing.flip:
            print(f"flipping image {preprocessing.flip.value}")

            flip_tform = gen_aff_tform_flip(
                image, image_res, preprocessing.flip.value
            )

            (
                composite_transform,
                _,
                final_tform,
            ) = prepare_wsireg_transform_data({"initial": [flip_tform]})

            image = transform_plane(image, final_tform, composite_transform)

            if self._mask is not None:
                self._mask.SetSpacing((image_res, image_res))
                self._mask = transform_plane(
                    self._mask, final_tform, composite_transform
                )

            transforms.append(flip_tform)

        if self._mask and preprocessing.crop_to_mask_bbox:
            print("computing mask bounding box")
            if preprocessing.mask_bbox is None:
                mask_bbox = compute_mask_to_bbox(self._mask)
                preprocessing.mask_bbox = mask_bbox

        if preprocessing.mask_bbox:

            print("cropping to mask")
            translation_transform = gen_rigid_translation(
                image,
                image_res,
                preprocessing.mask_bbox.X,
                preprocessing.mask_bbox.Y,
                preprocessing.mask_bbox.WIDTH,
                preprocessing.mask_bbox.HEIGHT,
            )

            (
                composite_transform,
                _,
                final_tform,
            ) = prepare_wsireg_transform_data(
                {"initial": [translation_transform]}
            )

            image = transform_plane(image, final_tform, composite_transform)

            self.original_size_transform = gen_rig_to_original(
                original_size, deepcopy(translation_transform)
            )

            if self._mask is not None:
                self._mask.SetSpacing((image_res, image_res))
                self._mask = transform_plane(
                    self._mask, final_tform, composite_transform
                )
            transforms.append(translation_transform)

        return image, transforms

    def preprocess_image(self, reg_image: sitk.Image) -> None:
        """
        Run full intensity and spatial preprocessing. Creates the `reg_image` attribute

        Parameters
        ----------
        reg_image: sitk.Image
            Raw form of image to be preprocessed

        """

        reg_image = self.preprocess_reg_image_intensity(
            reg_image, self.preprocessing
        )

        if reg_image.GetDepth() >= 1:
            raise ValueError(
                "preprocessing did not result in a single image plane\n"
                "multi-channel or 3D image return"
            )

        if reg_image.GetNumberOfComponentsPerPixel() > 1:
            raise ValueError(
                "preprocessing did not result in a single image plane\n"
                "multi-component / RGB(A) image returned"
            )

        reg_image, pre_reg_transforms = self.preprocess_reg_image_spatial(
            reg_image, self.preprocessing, self.pre_reg_transforms
        )

        if len(pre_reg_transforms) > 0:
            self.pre_reg_transforms = pre_reg_transforms

        self._reg_image = reg_image

    def reg_image_sitk_to_itk(self, cast_to_float32: bool = True) -> None:
        """
        Convert SimpleITK to ITK for use in ITKElastix.

        Parameters
        ----------
        cast_to_float32: bool
            Whether to make image float32 for ITK, needs to be true for registration.

        """
        origin = self._reg_image.GetOrigin()
        spacing = self._reg_image.GetSpacing()
        # direction = image.GetDirection()
        is_vector = self._reg_image.GetNumberOfComponentsPerPixel() > 1
        if cast_to_float32 is True:
            self._reg_image = sitk.Cast(self._reg_image, sitk.sitkFloat32)
            self._reg_image = sitk.GetArrayFromImage(self._reg_image)
        else:
            self._reg_image = sitk.GetArrayFromImage(self._reg_image)

        self._reg_image = itk.GetImageFromArray(
            self._reg_image, is_vector=is_vector
        )
        self._reg_image.SetOrigin(origin)
        self._reg_image.SetSpacing(spacing)

        if self._mask is not None:
            origin = self._mask.GetOrigin()
            spacing = self._mask.GetSpacing()
            # direction = image.GetDirection()
            is_vector = self._mask.GetNumberOfComponentsPerPixel() > 1
            if cast_to_float32 is True:
                self._mask = sitk.Cast(self._mask, sitk.sitkFloat32)
                self._mask = sitk.GetArrayFromImage(self._mask)
            else:
                self._mask = sitk.GetArrayFromImage(self._mask)

            self._mask = itk.GetImageFromArray(self._mask, is_vector=is_vector)
            self._mask.SetOrigin(origin)
            self._mask.SetSpacing(spacing)

            mask_im_type = itk.Image[itk.UC, 2]
            self._mask = itk.binary_threshold_image_filter(
                self._mask,
                lower_threshold=1,
                inside_value=1,
                ttype=(type(self._mask), mask_im_type),
            )

    @staticmethod
    def _get_all_cache_data_fps(output_dir: Union[str, Path], image_tag: str):
        """Get cached directories"""
        output_dir = Path(output_dir)

        out_image_fp = output_dir / f"{image_tag}_prepro.tiff"
        out_params_fp = output_dir / f"{image_tag}_preprocessing_params.json"
        out_mask_fp = output_dir / f"{image_tag}_prepro_mask.tiff"
        out_init_tform_fp = output_dir / f"{image_tag}_init_tforms.json"
        out_osize_tform_fp = output_dir / f"{image_tag}_orig_size_tform.json"

        return (
            out_image_fp,
            out_params_fp,
            out_mask_fp,
            out_init_tform_fp,
            out_osize_tform_fp,
        )

    def check_cache_preprocessing(
        self, output_dir: Union[str, Path], image_tag: str
    ):
        """

        Parameters
        ----------
        output_dir: path
            Where cached data is on disk
        image_tag:
            Tag of the image modality

        Returns
        -------
        prepro_flag: bool
            Whether a preprocessed version of the image exists in the cache.
        """
        (
            out_image_fp,
            out_params_fp,
            out_mask_fp,
            _,
            _,
        ) = self._get_all_cache_data_fps(output_dir, image_tag)

        if out_image_fp.exists() and out_params_fp.exists():
            cached_preprocessing = ImagePreproParams.parse_file(out_params_fp)
            return self.preprocessing == cached_preprocessing
        else:
            return False

    def cache_image_data(
        self, output_dir: Union[str, Path], image_tag: str, check: bool = True
    ) -> None:
        """
        Save preprocessed image data to a cache in WsiReg2D.
        Parameters
        ----------
        output_dir: path
            Where cached data is on disk
        image_tag:
            Tag of the image modality
        check: bool
            Whether to check for existence of data

        """

        (
            out_image_fp,
            out_params_fp,
            out_mask_fp,
            out_init_tform_fp,
            out_osize_tform_fp,
        ) = self._get_all_cache_data_fps(output_dir, image_tag)

        if check:
            read_from_cache = self.check_cache_preprocessing(
                output_dir, image_tag
            )
        else:
            read_from_cache = False

        if not read_from_cache:
            print(f"Writing preprocessed image for {image_tag}")
            sitk.WriteImage(
                self.reg_image, str(out_image_fp), useCompression=True
            )
            print(f"Finished writing preprocessed image for {image_tag}")
            json.dump(
                deepcopy(
                    self.preprocessing.dict(
                        exclude_none=True, exclude_defaults=True
                    )
                ),
                open(out_params_fp, "w"),
                cls=NpEncoder,
            )
            json.dump(self.pre_reg_transforms, open(out_init_tform_fp, "w"))

            if self._mask is not None:
                print(f"Writing preprocessed mask for {image_tag}")
                sitk.WriteImage(
                    self.mask, str(out_mask_fp), useCompression=True
                )
                print(f"Finished writing preprocessed mask for {image_tag}")

            if self.original_size_transform:
                json.dump(
                    self.original_size_transform, open(out_osize_tform_fp, "w")
                )

    def load_from_cache(self, output_dir: Union[str, Path], image_tag: str):
        """
        Read in preprocessed data from the cache folder.
        Parameters
        ----------
        output_dir: path
            Where cached data is on disk
        image_tag:
            Tag of the image modality

        Returns
        -------
        from_cache_flag: bool
            Whether data was read from cache
        """
        (
            image_fp,
            params_fp,
            mask_fp,
            init_tform_fp,
            osize_tform_fp,
        ) = self._get_all_cache_data_fps(output_dir, image_tag)

        read_from_cache = self.check_cache_preprocessing(output_dir, image_tag)

        if read_from_cache:
            self._reg_image = sitk.ReadImage(str(image_fp))
            self._preprocessing = ImagePreproParams(
                **json.load(open(params_fp, "r"))
            )
            self.pre_reg_transforms = json.load(open(init_tform_fp, "r"))
            if osize_tform_fp.exists():
                self.original_size_transform = json.load(
                    open(osize_tform_fp, "r")
                )

            if mask_fp.exists():
                self._mask = sitk.ReadImage(str(mask_fp))
            return True
        else:
            return False

    @staticmethod
    def load_orignal_size_transform(
        output_dir: Union[str, Path], image_tag: str
    ):
        """
        Read original size transform from cache.

        Parameters
        ----------
        output_dir: path
            Where cached data is on disk
        image_tag:
            Tag of the image modality

        Returns
        -------
        osize_tform: list
            Original size transform or empty

        """
        (
            _,
            _,
            _,
            init_tform_fp,
            _,
        ) = RegImage._get_all_cache_data_fps(output_dir, image_tag)

        if init_tform_fp.exists():
            return [json.load(open(init_tform_fp, "r"))]
        else:
            return []


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)



import warnings

import SimpleITK as sitk

from wsireg.reg_images.reg_image import RegImage
from wsireg.utils.im_utils import (
    ensure_dask_array,
    get_sitk_image_info,
    guess_rgb,
)


class SitkRegImage(RegImage):
    def __init__(
        self,
        image,
        image_res,
        mask=None,
        pre_reg_transforms=None,
        preprocessing=None,
        channel_names=None,
        channel_colors=None,
    ):
        super(SitkRegImage, self).__init__(preprocessing)
        self._path = image
        self._image_res = image_res

        self.reader = "sitk"

        (
            self._shape,
            self._im_dtype,
        ) = self._get_image_info()

        self._is_rgb = guess_rgb(self._shape)

        self._n_ch = self._shape[2] if self.is_rgb else self._shape[0]

        if mask:
            self._mask = self.read_mask(mask)

        self.pre_reg_transforms = pre_reg_transforms

        self._channel_names = channel_names
        self._channel_colors = channel_colors
        self.original_size_transform = None

    def _get_image_info(self):

        im_dims, im_dtype = get_sitk_image_info(self._path)

        im_dims = (int(im_dims[0]), int(im_dims[1]), int(im_dims[2]))

        return im_dims, im_dtype

    def read_reg_image(self):
        """
        Read and preprocess the image for registration.
        """
        reg_image = sitk.ReadImage(self._path)

        if self.preprocessing.as_uint8 is True and reg_image.GetPixelID() != 1:
            reg_image = sitk.Cast(
                sitk.RescaleIntensity(reg_image), sitk.sitkUInt8
            )

        self.preprocess_image(reg_image)

    def _read_full_image(self):
        self._dask_image = ensure_dask_array(
            sitk.GetArrayFromImage(sitk.ReadImage(self._path))
        )

        rechunk_size = (
            (2048, 2048, self.n_ch) if self.is_rgb else (self.n_ch, 2048, 2048)
        )
        self._dask_image = self._dask_image.rechunk(rechunk_size)

    def read_single_channel(self, channel_idx: int):
        """
        Read in a single channel for transformation by plane.
        Parameters
        ----------
        channel_idx: int
            Index of the channel to be read

        Returns
        -------
        image: np.ndarray
            Numpy array of the selected channel to be read
        """
        if channel_idx > (self.n_ch - 1):
            warnings.warn(
                "channel_idx exceeds number of channels, reading channel at channel_idx == 0"
            )
            channel_idx = 0

        if self._is_rgb:
            image = self._dask_image[:, :, channel_idx].compute()
        else:
            image = self._dask_image[channel_idx, :, :].compute()

        return image



import warnings
from typing import Tuple

import dask.array as da
import numpy as np
import SimpleITK as sitk

from wsireg.reg_images.reg_image import RegImage
from wsireg.utils.im_utils import CziRegImageReader, guess_rgb


class CziRegImage(RegImage):
    def __init__(
        self,
        image,
        image_res,
        mask=None,
        pre_reg_transforms=None,
        preprocessing=None,
        channel_names=None,
        channel_colors=None,
    ):
        super(CziRegImage, self).__init__(preprocessing)
        self._path = image
        self._image_res = image_res

        self.czi = CziRegImageReader(self._path)
        self.reader = "czi"

        scene_idx = self.czi.axes.index('S')

        if self.czi.shape[scene_idx] > 1:
            raise ValueError('multi scene czis not allowed at this time')

        (
            self._shape,
            self._im_dtype,
        ) = self._get_image_info()

        self._is_rgb = guess_rgb(self._shape)
        self._n_ch = self._shape[2] if self.is_rgb else self._shape[0]

        self._dask_image = self._prepare_dask_image()

        if mask:
            self._mask = self.read_mask(mask)

        self.pre_reg_transforms = pre_reg_transforms

        self._channel_names = channel_names
        self._channel_colors = channel_colors
        self.original_size_transform = None

    def _get_image_info(self):
        # if RGB need to get 0
        if self.czi.shape[-1] > 1:
            ch_dim_idx = self.czi.axes.index('0')
        else:
            ch_dim_idx = self.czi.axes.index('C')
        y_dim_idx = self.czi.axes.index('Y')
        x_dim_idx = self.czi.axes.index('X')
        if self.czi.shape[-1] > 1:
            im_dims = np.array(self.czi.shape)[
                [y_dim_idx, x_dim_idx, ch_dim_idx]
            ]
        else:
            im_dims = np.array(self.czi.shape)[
                [ch_dim_idx, y_dim_idx, x_dim_idx]
            ]

        im_dtype = self.czi.dtype

        im_dims = (int(im_dims[0]), int(im_dims[1]), int(im_dims[2]))

        return im_dims, im_dtype

    def _prepare_dask_image(self) -> da.Array:
        ch_dim = self._shape[1:] if not self._is_rgb else self._shape[:2]
        chunks = ((1,) * self._n_ch, (ch_dim[0],), (ch_dim[1],))
        dask_image = da.map_blocks(
            self._czi_read_single_channel,
            chunks=chunks,
            dtype=self.im_dtype,
            meta=np.array((), dtype=self._im_dtype),
        )
        return dask_image

    def _czi_read_single_channel(self, block_id: Tuple[int, ...]):
        channel_idx = block_id[0]
        if self.is_rgb is False:
            image = self.czi.sub_asarray(
                channel_idx=[channel_idx],
            )
        else:
            image = self.czi.sub_asarray_rgb(
                channel_idx=[channel_idx], greyscale=False
            )

        return np.expand_dims(np.squeeze(image), axis=0)

    def read_reg_image(self):
        """
        Read and preprocess the image for registration.
        For the Zeiss CZI reader, this involves grayscaling RGB on read
        or reading only a subset of the channel images.
        """
        if self.is_rgb:
            reg_image = self.czi.sub_asarray_rgb(greyscale=True)
        else:
            reg_image = self.czi.sub_asarray(
                channel_idx=self.preprocessing.ch_indices,
                as_uint8=self.preprocessing.as_uint8,
            )

        reg_image = np.squeeze(reg_image)
        reg_image = sitk.GetImageFromArray(reg_image)

        self.preprocess_image(reg_image)

    def read_single_channel(self, channel_idx: int):
        """
        Read in a single channel for transformation by plane.
        Parameters
        ----------
        channel_idx: int
            Index of the channel to be read

        Returns
        -------
        image: np.ndarray
            Numpy array of the selected channel to be read
        """
        if channel_idx > (self.n_ch - 1):
            warnings.warn(
                "channel_idx exceeds number of channels, reading channel at channel_idx == 0"
            )
            channel_idx = 0

        image = self._dask_image[channel_idx, :, :].compute()

        return image



from pathlib import Path
from typing import List, Optional, Union
from warnings import warn

import numpy as np

from wsireg.reg_images.loader import reg_image_loader


class MergeRegImage:
    def __init__(
        self,
        image_fp: List[Union[Path, str]],
        image_res: List[Union[int, float]],
        channel_names: Optional[List[List[str]]] = None,
        channel_colors: Optional[List[List[str]]] = None,
    ):

        if isinstance(image_fp, list) is False:
            raise ValueError(
                "MergeRegImage requires a list of images to merge"
            )

        if isinstance(image_res, list) is False:
            raise ValueError(
                "MergeRegImage requires a list of image resolutions for each image to merge"
            )

        if channel_names is None:
            channel_names = [None for _ in range(0, len(image_fp))]

        if channel_colors is None:
            channel_colors = [None for _ in range(0, len(image_fp))]

        images = []
        for im_idx, image_data in enumerate(
            zip(image_fp, image_res, channel_names, channel_colors)
        ):
            image, image_res, channel_names, channel_colors = image_data
            imdata = reg_image_loader(
                image,
                image_res,
                channel_names=channel_names,
                channel_colors=channel_colors,
            )
            if (
                imdata.channel_names is None
                or len(imdata.channel_names) != imdata.n_ch
            ):
                imdata._channel_names = [
                    f"C{idx}" for idx in range(0, imdata.n_ch)
                ]

            images.append(imdata)

        if all([im.im_dtype == images[0].im_dtype for im in images]) is False:
            warn(
                "MergeRegImage created with mixed data types, writing will cast "
                "to the largest data type"
            )

        if any([im.is_rgb for im in images]) is True:
            warn(
                "MergeRegImage does not support writing merged interleaved RGB "
                "Data will be written as multi-channel"
            )

        self.images = images
        self.image_fps = image_fp
        self.im_dtype = self.images[0].im_dtype

        self.is_rgb = False

        self.n_ch = np.sum([i.n_ch for i in self.images])
        self.channel_names = [i.channel_names for i in self.images]
        self.original_size_transform = None



import warnings
from typing import List, Tuple

import dask.array as da
import numpy as np
import SimpleITK as sitk
from ome_types import from_xml
from tifffile import TiffFile

from wsireg.reg_images.reg_image import RegImage
from wsireg.utils.im_utils import (
    get_tifffile_info,
    guess_rgb,
    preprocess_dask_array,
    tifffile_to_dask,
)


class TiffFileRegImage(RegImage):
    def __init__(
        self,
        image_fp,
        image_res,
        mask=None,
        pre_reg_transforms=None,
        preprocessing=None,
        channel_names=None,
        channel_colors=None,
    ):
        super(TiffFileRegImage, self).__init__(preprocessing)
        self._path = image_fp
        self._image_res = image_res
        self.tf = TiffFile(self._path)
        self.reader = "tifffile"

        (
            self._shape,
            self._im_dtype,
            self.largest_series,
        ) = self._get_image_info()

        self._get_dim_info()

        self._dask_image = self._get_dask_image()

        if mask:
            self._mask = self.read_mask(mask)

        self.pre_reg_transforms = pre_reg_transforms

        self._channel_names = channel_names
        self._channel_colors = channel_colors
        self.original_size_transform = None

    def _get_image_info(self) -> Tuple[Tuple[int, int, int], np.dtype, int]:
        if len(self.tf.series) > 1:
            warnings.warn(
                "The tiff contains multiple series, "
                "the largest series will be read by default"
            )

        im_dims, im_dtype, largest_series = get_tifffile_info(self._path)

        im_dims = (int(im_dims[0]), int(im_dims[1]), int(im_dims[2]))

        return im_dims, im_dtype, largest_series

    def _get_dim_info(self) -> None:
        if self._shape:
            if self.tf.ome_metadata:
                self.ome_metadata = from_xml(self.tf.ome_metadata)
                spp = (
                    self.ome_metadata.images[self.largest_series]
                    .pixels.channels[0]
                    .samples_per_pixel
                )
                interleaved = self.ome_metadata.images[
                    self.largest_series
                ].pixels.interleaved

                if spp and spp > 1:
                    self._is_rgb = True
                else:
                    self._is_rgb = False

                if guess_rgb(self._shape) is False:
                    self._channel_axis = 0
                    self._is_interleaved = False
                elif interleaved and guess_rgb(self._shape):
                    self._is_interleaved = True
                    self._channel_axis = len(self._shape) - 1

            else:
                self._is_rgb = guess_rgb(self._shape)
                self._is_interleaved = self._is_rgb
                if self._is_rgb:
                    self._channel_axis = len(self._shape) - 1
                else:
                    self._channel_axis = 0

            self._n_ch = self._shape[self._channel_axis]

    def _get_dask_image(self) -> List[da.Array]:
        dask_image = tifffile_to_dask(self._path, self.largest_series, level=0)
        dask_image = (
            dask_image.reshape(1, *dask_image.shape)
            if len(dask_image.shape) == 2
            else dask_image
        )

        if self._is_rgb and not self._is_interleaved:
            dask_image = da.rollaxis(dask_image, 0, 3)

        return dask_image

    def read_reg_image(self):
        """
        Read and preprocess the image for registration.
        """
        reg_image = self._dask_image
        reg_image = preprocess_dask_array(reg_image, self.preprocessing)

        if (
            self.preprocessing is not None
            and self.preprocessing.as_uint8 is True
            and reg_image.GetPixelID() != sitk.sitkUInt8
        ):
            reg_image = sitk.RescaleIntensity(reg_image)
            reg_image = sitk.Cast(reg_image, sitk.sitkUInt8)

        self.preprocess_image(reg_image)

    def read_single_channel(self, channel_idx: int):
        """
        Read in a single channel for transformation by plane.
        Parameters
        ----------
        channel_idx: int
            Index of the channel to be read

        Returns
        -------
        image: np.ndarray
            Numpy array of the selected channel to be read
        """
        if channel_idx > (self.n_ch - 1):
            warnings.warn(
                "channel_idx exceeds number of channels, reading channel at channel_idx == 0"
            )
            channel_idx = 0
        if self._is_rgb:
            image = self._dask_image[:, :, channel_idx].compute()
        else:
            image = self._dask_image[channel_idx, :, :].compute()

        return image



from pathlib import Path
import os
import random
import string

import numpy as np
import pytest
from tifffile import imread
import dask.array as da

from wsireg.reg_images.loader import reg_image_loader
from wsireg.reg_images.merge_reg_image import MergeRegImage
from wsireg.reg_transforms.reg_transform_seq import RegTransformSeq
from wsireg.writers.merge_ome_tiff_writer import MergeOmeTiffWriter
from wsireg.writers.ome_tiff_writer import OmeTiffWriter
from wsireg.writers.tiled_ome_tiff_writer import OmeTiffTiledWriter

HERE = os.path.dirname(__file__)
TFORM_FP = os.path.join(HERE, "fixtures/complex_linear_reg_transform.json")


def gen_project_name_str():
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(10))


@pytest.mark.usefixtures("complex_transform")
def test_OmeTiffWriter_by_plane(complex_transform, tmp_path):
    reg_image = reg_image_loader(np.ones((1024, 1024), dtype=np.uint8), 1)
    # composite_transform, _, final_transform = prepare_wsireg_transform_data(
    #     complex_transform
    # )
    rts = RegTransformSeq(complex_transform)

    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)
    by_plane_fp = ometiffwriter.write_image_by_plane(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )
    by_plane_image = reg_image_loader(by_plane_fp, 2)
    assert by_plane_image.shape == (1, 1024, 1024)


@pytest.mark.usefixtures("complex_transform")
def test_OmeTiffWriter_by_tile(complex_transform, tmp_path):
    reg_image = reg_image_loader(np.ones((4096, 4096), dtype=np.uint8), 0.5)
    rts = RegTransformSeq(complex_transform)

    ometiffwriter = OmeTiffTiledWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiffwriter.write_image_by_tile(
        gen_project_name_str(),
        output_dir=str(tmp_path),
        zarr_temp_dir=tmp_path,
    )
    by_tile_image = reg_image_loader(by_tile_fp, 2)
    assert by_tile_image.shape == (1, 1024, 1024)


@pytest.mark.usefixtures("simple_transform_affine")
def test_OmeTiffWriter_compare_tile_plane(simple_transform_affine, tmp_path):
    reg_image = reg_image_loader(np.ones((1024, 1024), dtype=np.uint8), 1)
    rts = RegTransformSeq(simple_transform_affine)
    ometifftilewriter = OmeTiffTiledWriter(reg_image, reg_transform_seq=rts)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometifftilewriter.write_image_by_tile(
        gen_project_name_str(),
        output_dir=str(tmp_path),
        zarr_temp_dir=tmp_path,
    )
    by_tile_image = reg_image_loader(by_tile_fp, 2)

    by_plane_fp = ometiffwriter.write_image_by_plane(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )
    by_plane_image = reg_image_loader(by_plane_fp, 2)

    ch0_im_tile = by_tile_image.read_single_channel(0)
    ch0_im_plane = by_plane_image.read_single_channel(0)

    assert by_plane_image.shape == by_tile_image.shape
    assert np.array_equal(ch0_im_tile, ch0_im_plane)


@pytest.mark.usefixtures("simple_transform_affine")
def test_OmeTiffWriter_compare_tile_plane_rgb(
    simple_transform_affine, tmp_path
):
    reg_image = reg_image_loader(
        np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint8), 1
    )

    rts = RegTransformSeq(simple_transform_affine)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)
    ometiletiffwriter = OmeTiffTiledWriter(reg_image, reg_transform_seq=rts)
    by_tile_fp = ometiletiffwriter.write_image_by_tile(
        gen_project_name_str(),
        output_dir=str(tmp_path),
        zarr_temp_dir=tmp_path,
    )

    by_plane_fp = ometiffwriter.write_image_by_plane(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )
    Path(by_tile_fp).as_posix()
    im_tile = imread(by_tile_fp)
    im_plane = imread(by_plane_fp)
    assert np.array_equal(im_tile, im_plane)


@pytest.mark.usefixtures("simple_transform_affine_nl")
def test_OmeTiffWriter_compare_tile_plane_rgb_nl(
    simple_transform_affine_nl, tmp_path
):
    reg_image = reg_image_loader(
        np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint8), 1
    )

    rts = RegTransformSeq(simple_transform_affine_nl)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)
    ometiletiffwriter = OmeTiffTiledWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiletiffwriter.write_image_by_tile(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )

    by_plane_fp = ometiffwriter.write_image_by_plane(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )

    im_tile = imread(by_tile_fp)
    im_plane = imread(by_plane_fp)

    assert np.array_equal(im_tile, im_plane)


@pytest.mark.usefixtures("simple_transform_affine")
def test_OmeTiffWriter_compare_tile_plane_mc(
    simple_transform_affine, tmp_path
):
    reg_image = reg_image_loader(
        np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8), 1
    )
    rts = RegTransformSeq(simple_transform_affine)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)
    ometiletiffwriter = OmeTiffTiledWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiletiffwriter.write_image_by_tile(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )

    by_plane_fp = ometiffwriter.write_image_by_plane(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )

    im_tile = imread(by_tile_fp)
    im_plane = imread(by_plane_fp)

    assert np.array_equal(im_tile, im_plane)


@pytest.mark.usefixtures("simple_transform_affine_nl")
def test_OmeTiffWriter_compare_tile_plane_mc_nl(
    simple_transform_affine_nl, tmp_path
):
    reg_image = reg_image_loader(
        np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8), 1
    )
    rts = RegTransformSeq(simple_transform_affine_nl)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)
    ometiletiffwriter = OmeTiffTiledWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiletiffwriter.write_image_by_tile(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )

    by_plane_fp = ometiffwriter.write_image_by_plane(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )

    im_tile = imread(by_tile_fp)
    im_plane = imread(by_plane_fp)

    assert np.array_equal(im_tile, im_plane)


def test_OmeTiffWriter_compare_tile_plane_mc_nl_large(tmp_path):
    im_array = da.from_array(
        np.random.randint(0, 255, (2, 2**13, 2**13), dtype=np.uint8),
        chunks=(1, 1024, 1024),
    )
    reg_image = reg_image_loader(im_array, 0.5)

    rts = RegTransformSeq(TFORM_FP)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)
    ometiletiffwriter = OmeTiffTiledWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiletiffwriter.write_image_by_tile(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )

    by_plane_fp = ometiffwriter.write_image_by_plane(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )

    im_tile = imread(by_tile_fp)
    im_plane = imread(by_plane_fp)

    assert np.array_equal(im_tile, im_plane)


def test_OmeTiffWriter_compare_tile_plane_rgb_nl_large(tmp_path):
    im_array = da.from_array(
        np.random.randint(0, 255, (2**13, 2**13, 3), dtype=np.uint8),
        chunks=(1024, 1024, 3),
    )
    reg_image = reg_image_loader(im_array, 0.5)

    rts = RegTransformSeq(TFORM_FP)
    ometiffwriter = OmeTiffWriter(reg_image, reg_transform_seq=rts)
    ometiletiffwriter = OmeTiffTiledWriter(reg_image, reg_transform_seq=rts)

    by_tile_fp = ometiletiffwriter.write_image_by_tile(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )

    by_plane_fp = ometiffwriter.write_image_by_plane(
        gen_project_name_str(),
        output_dir=str(tmp_path),
    )

    im_tile = imread(by_tile_fp)
    im_plane = imread(by_plane_fp)

    assert np.array_equal(im_tile, im_plane)


@pytest.mark.usefixtures("simple_transform_affine_nl")
def test_MergeOmeTiffWriter_mc(simple_transform_affine_nl, tmp_path):
    reg_image1 = np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8)
    reg_image2 = np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8)
    reg_image3 = np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8)

    mreg_image = MergeRegImage(
        [reg_image1, reg_image2, reg_image3],
        [1, 1, 1],
        channel_names=[["1", "2", "3"], ["1", "2", "3"], ["1", "2", "3"]],
    )
    rts = RegTransformSeq(simple_transform_affine_nl)
    merge_ometiffwriter = MergeOmeTiffWriter(
        mreg_image, reg_transform_seqs=[rts, rts, rts]
    )

    by_plane_fp = merge_ometiffwriter.merge_write_image_by_plane(
        "merge_testimage_by_plane",
        ["1", "2", "3"],
        output_dir=str(tmp_path),
    )

    im_plane = imread(by_plane_fp)

    reg_image1_loaded = reg_image_loader(reg_image1, 1)

    ometiffwriter = OmeTiffWriter(reg_image1_loaded, reg_transform_seq=rts)

    by_plane_fp_s1 = ometiffwriter.write_image_by_plane(
        "testimage_by_plane_s1",
        output_dir=str(tmp_path),
    )

    im_plane_s1 = imread(by_plane_fp_s1)

    reg_image2_loaded = reg_image_loader(reg_image2, 1)

    ometiffwriter = OmeTiffWriter(reg_image2_loaded, reg_transform_seq=rts)

    by_plane_fp_s2 = ometiffwriter.write_image_by_plane(
        "testimage_by_plane_s2",
        output_dir=str(tmp_path),
    )

    im_plane_s2 = imread(by_plane_fp_s2)

    reg_image3_loaded = reg_image_loader(reg_image3, 1)

    ometiffwriter = OmeTiffWriter(reg_image3_loaded, reg_transform_seq=rts)

    by_plane_fp_s3 = ometiffwriter.write_image_by_plane(
        "testimage_by_plane_s3",
        output_dir=str(tmp_path),
    )

    im_plane_s3 = imread(by_plane_fp_s3)

    assert im_plane.shape[0] == 9
    assert np.array_equal(im_plane[0:3, :, :], im_plane_s1)
    assert np.array_equal(im_plane[3:6, :, :], im_plane_s2)
    assert np.array_equal(im_plane[6:9, :, :], im_plane_s3)


@pytest.mark.usefixtures("simple_transform_affine_nl")
def test_MergeOmeTiffWriter_mix_merge(simple_transform_affine_nl, tmp_path):

    reg_image1 = np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint16)
    reg_image2 = np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint16)
    reg_image3 = np.random.randint(0, 255, (3, 1024, 1024), dtype=np.uint8)

    mreg_image = MergeRegImage(
        [reg_image1, reg_image2, reg_image3],
        [1, 1, 1],
        channel_names=[["1", "2", "3"], ["1", "2", "3"], ["1", "2", "3"]],
    )

    rts = RegTransformSeq(simple_transform_affine_nl)
    merge_ometiffwriter = MergeOmeTiffWriter(
        mreg_image, reg_transform_seqs=[rts, rts, rts]
    )

    by_plane_fp = merge_ometiffwriter.merge_write_image_by_plane(
        "merge_testimage_by_plane_mix",
        ["1", "2", "3"],
        output_dir=str(tmp_path),
    )

    im_plane = imread(by_plane_fp)

    assert im_plane.shape[0] == 9
